/*********************************************************************
  ulib/graphics/lowlevel/img2raw.hh

  purpose:
    Classes that allow image data stored in Image_*<Pixel> objects
    in either RGB or YUV format to be converted to a format the
    hardware directly understands. This can be (for example) the
    data ordering required for X11 XImages.

    Special hardware oriented convertion routines should be
    integrated here. The appropriate transformation is selected
    automagically.

  notes:

  to do:

  author(s):
   - Dirk Farin, farin@ti.uni-mannheim.de
     University Mannheim, Dept. Circuitry and Simulation
     B 6,26 EG, room 0.10 / D-68131 Mannheim / Germany

  modifications:
    29/Jul/1999 - Dirk Farin - first implementation
 *********************************************************************/

#ifndef ULIB_GRAPHICS_LOWLEVEL_IMG2RAW_HH
#define ULIB_GRAPHICS_LOWLEVEL_IMG2RAW_HH

#include "ulib/graphics/basic/image.hh"
#include "ulib/types.hh"

class Image2Raw_TransformRGB
{
public:
  virtual ~Image2Raw_TransformRGB() { }
  void SetBytesPerLine(int bpl) { d_bpl=bpl; }
  virtual void Transform(const Image_RGB<Pixel>&,uint8* mem) { assert(0); }

  virtual const char* TransformationName() = 0;

protected:
  int d_bpl;
};

class Image2Raw_TransformYUV
{
public:
  virtual ~Image2Raw_TransformYUV() { }
  void SetBytesPerLine(int bpl) { d_bpl=bpl; }
  virtual void Transform(const Image_YUV<Pixel>&,uint8* mem) { assert(0); }

  virtual const char* TransformationName() = 0;

protected:
  int d_bpl;
};


class Image2Raw
{
public:
  Image2Raw();
  ~Image2Raw();

  void InputIsRGB(bool flag=true) { d_rgbinput=flag; }
  void InputIsGreyscale(bool flag=true) { d_greyscale=flag; }
  void SetInputChromaFormat(ChromaFormat f) { d_chroma=f; }
  void SetOutputMem(uint8* mem);
  void SetOutputParam(int bytes_per_line,int bits_per_pixel,bool littleendian,
		      uint32 redmask,uint32 greenmask,uint32 bluemask);

  void SetZoomFactor(int f=2) { assert(0); /* TODO */ }
  void SetGrayscale(bool flag=true) { assert(0); /* TODO */ }

  void TransformRGB(const Image_RGB<Pixel>&);
  void TransformYUV(const Image_YUV<Pixel>&);

private:
  uint8* d_mem;
  int    d_bytes_per_line;
  int    d_bits_per_pixel;
  bool   d_littleendian;
  uint32 d_redmask;
  uint32 d_greenmask;
  uint32 d_bluemask;
  bool   d_rgbinput;
  bool   d_greyscale;
  ChromaFormat d_chroma;

  int red_bits  ,red_shift;
  int green_bits,green_shift;
  int blue_bits ,blue_shift;

  union {
    class Image2Raw_TransformRGB* d_rgbtransform;
    class Image2Raw_TransformYUV* d_yuvtransform;
  };
};

void CalcBitsShift(uint32 mask,int& bits,int& shift);

#endif
